from flask import Blueprint, request
from flask_jwt_extended import jwt_required, get_jwt_identity
from flask_restful import Api, Resource

from portal.rest import submission_schema, submission_state_schema, review_schema, reviews_schema
from portal.tools.decorators import error_handler
from portal.tools.logging import log
from portal.service import service
from portal.service.errors import PortalAPIError, ForbiddenError
from portal.service import policies

submissions = Blueprint('submissions', __name__)
submissions_api = Api(submissions)


class SubmissionResource(Resource):
    @error_handler
    @jwt_required
    def get(self, sid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        submission = service.find_submission(sid)
        # authorization
        if policies.check_component(component=client, course=submission.project.course, permissions=['readSubmissions']):
            pass
        elif policies.check_user(user=client, course=submission.project.course, permissions=['read_submissions_all']):
            pass
        elif policies.check_user(user=client, course=submission.project.course, permissions=['read_submissions_groups']):
            # TODO
            pass
        elif policies.check_user(user=client, course=submission.project.course, permissions=['read_submissions_own']) \
                and submission.user == client:
            pass
        else:
            raise ForbiddenError(uid=client.id)
        return submission_schema.dump(submission)

    @error_handler
    @jwt_required
    def delete(self, sid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        submission = service.find_submission(sid)
        # authorization
        if not policies.check_sysadmin(client):
            raise ForbiddenError(uid=client)

        service.delete_entity(submission)
        return '', 204


class SubmissionState(Resource):
    @error_handler
    @jwt_required
    def get(self, sid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        submission = service.find_submission(sid)
        # authorization
        if policies.check_component(component=client, permissions=['readSubmissionState'], course=submission.project.course):
            pass
        elif policies.check_user(user=client, course=submission.project.course, permissions=['read_submissions_all']):
            pass
        elif policies.check_user(user=client, course=submission.project.course, permissions=['read_submissions_groups']):
            # TODO
            pass
        elif policies.check_user(user=client, course=submission.project.course, permissions=['read_submissions_own']) \
                and submission.user == client:
            pass
        else:
            raise ForbiddenError(uid=client)

        return submission_state_schema.dump(submission)[0], 200

    @error_handler
    @jwt_required
    def put(self, sid):
        log.info(f"PUT to {request.url}")
        client = service.find_client()
        submission = service.find_submission(sid)
        # authorization
        full = False
        if client.id == submission.user_id:
            pass
        elif (policies.check_component(component=client, permissions=['writeSubmissionState'],
                                       course=submission.project.course)
              or policies.check_sysadmin(client)):
            full = True
        else:
            raise ForbiddenError(uid=client)

        json_data = request.get_json()
        if not json_data:
            raise PortalAPIError(400, message='No data provided for submission state update.')

        data = submission_state_schema.load(json_data)[0]
        new_state = data['state']
        from database import SubmissionState
        if new_state != SubmissionState.CANCELLED and not full:
            raise ForbiddenError(uid=client.id, note=f"User {client.id} cannot update state to other than CANCELLED.")

        submission.change_state(new_state)
        service.write_entity(submission)
        log.info(f"Updated state of submission {sid} to {submission.state}")
        return '', 204


class SubmissionResult(Resource):
    @error_handler
    @jwt_required
    def put(self, sid):
        log.info(f"PUT to {request.url}")
        client = service.find_client()
        submission = service.find_submission(sid)
        # authorization

        # TODO load zip, save to storage


class SubmissionFiles(Resource):
    @error_handler
    @jwt_required
    def get(self, sid):
        client = service.find_client()
        # the user wants a zip or a json
        # authorization

        pass


class SubmissionSourceFiles(Resource):
    @error_handler
    @jwt_required
    def get(self, sid):
        client = service.find_client()
        pass


class SubmissionResultFiles(Resource):
    @error_handler
    @jwt_required
    def get(self, sid):
        client = service.find_client()
        pass


class SubmissionTestFiles(Resource):
    @error_handler
    @jwt_required
    def get(self, sid):
        client = service.find_client()
        pass


class SubmissionSources(Resource):
    @error_handler
    @jwt_required
    def get(self, sid):
        client = service.find_client()
        pass


class SubmissionResubmit(Resource):
    @error_handler
    @jwt_required
    def post(self, sid):
        log.info(f"POST to {request.url}")
        client = service.find_client()
        source_submission = service.find_submission(sid)
        # authorization
        if not (policies.check_component(component=client, permissions=['resubmit_submissions'],
                                         course=source_submission.project.course)
                or policies.check_user(user=client, course=source_submission.project.course,
                                       permissions=['resubmit_submissions'])):
            raise ForbiddenError(uid=client)

        json_data = request.get_json()
        if not json_data:
            raise PortalAPIError(400, message='No data provided for submission resubmit.')

        data = submission_schema.load(json_data)[0]
        # create new submission by copying files from the source submission in storage
        new_submission = service.copy_submission(source_submission, note=data['note'])
        return submission_schema.dump(new_submission), 201


class SubmissionReview(Resource):
    @error_handler
    @jwt_required
    def get(self, sid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        submission = service.find_submission(sid)
        # authorization
        if policies.check_component(component=client, permissions=['readSubmissionReview'], course=submission.project.course):
            pass
        elif policies.check_user(user=client, course=submission.project.course, permissions=['read_reviews_all']):
            pass
        elif policies.check_user(user=client, course=submission.project.course, permissions=['readGroupReviews']):
            # TODO
            pass
        elif policies.check_user(user=client, course=submission.project.course, permissions=['read_reviews_own']) \
                and submission.user == client:
            pass
        else:
            raise ForbiddenError(uid=client)

        return review_schema.dump(submission.review)[0], 200

    @error_handler
    @jwt_required
    def post(self, sid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        submission = service.find_submission(sid)
        # authorization
        if policies.check_component(component=client, permissions=['writeSubmissionReview'], course=submission.project.course):
            pass
        elif policies.check_user(user=client, course=submission.project.course, permissions=['writeOwnReview']):
            pass
        elif policies.check_user(user=client, course=submission.project.course, permissions=['write_reviews_group']):
            # TODO
            pass
        elif policies.check_user(user=client, course=submission.project.course, permissions=['write_reviews_own']) \
                and submission.user == client:
            pass
        else:
            raise ForbiddenError(uid=client)

        if not submission.review:
            service.create_review(submission)
        json_data = request.get_json()
        if not json_data:
            raise PortalAPIError(400, message='No data provided for group creation.')

        data = review_schema.load(json_data)[0]

        # also writes to db
        service.create_review_items(review=submission.review, items=data['review_items'], author=client)
        log.info(f"Added review items {data['review_items']} to review {submission.review.id} "
                 f"for submission {submission.id}")
        return review_schema.dump(submission.review)[0], 201


submissions_api.add_resource(SubmissionResource, '/<string:sid>')
submissions_api.add_resource(SubmissionState, '/<string:sid>/state')
submissions_api.add_resource(SubmissionResult, '/<string:sid>/result')
submissions_api.add_resource(SubmissionResubmit, '/<string:sid>/resubmit')

submissions_api.add_resource(SubmissionFiles, '/<string:sid>/files')
submissions_api.add_resource(SubmissionSourceFiles, '/<string:sid>/files/sources')
submissions_api.add_resource(SubmissionResultFiles, '/<string:sid>/files/results')

submissions_api.add_resource(SubmissionTestFiles, '/<string:sid>/test_files')
submissions_api.add_resource(SubmissionSources, '/<string:sid>/sources')

submissions_api.add_resource(SubmissionReview, '/<string:sid>/review')
