from flask import Blueprint, request, json
from flask_jwt_extended import jwt_required
from flask_restful import Api, Resource

from portal.rest import user_schema, users_schema, submissions_schema, roles_schema, courses_schema, groups_schema, \
    reviews_schema, user_schema_reduced, password_change_schema
from portal.service.errors import PortalAPIError, ForbiddenError, IncorrectPasswordError
from portal.service import service, policies
from portal.database.models import User
from portal.tools.logging import log
from portal.tools.decorators import error_handler
from portal.tools.time import DateTimeEncoder

users = Blueprint('users', __name__)
users_api = Api(users)


class UserResource(Resource):
    @error_handler
    @jwt_required
    def get(self, uid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        user = service.find_user(uid)
        # authorization
        if policies.check_component(component=client, permissions=[]) \
                or policies.check_sysadmin(client) \
                or client == user:
            return user_schema.dump(user)[0], 200
        return user_schema_reduced.dump(user)[0], 200

    @error_handler
    @jwt_required
    def put(self, uid):
        log.info(f"PUT to {request.url}")
        client = service.find_client()
        user = service.find_user(uid)
        # authorization
        if not (policies.check_component(component=client, permissions=[])
                or policies.check_sysadmin(client)
                or client == user):
            raise ForbiddenError(uid=client.id)
        full = False
        if policies.check_sysadmin(client):
            full = True

        log.debug(f'authorized for full user profile change: {full}')
        json_data = request.get_json()
        if not json_data:
            raise PortalAPIError(400, message='No data provided for user update.')
        data = user_schema.load(json_data)[0]
        log.warn(data)
        user.email = data['email']
        user.uco = data['uco']
        if data.get('name') is not None:
            user.name = data['name']
        if (data.get("is_admin") is not None) and full:
            user.is_admin = data['is_admin']

        log.debug(f"updating user: {user}")
        service.write_entity(user)
        log.debug(f"Updated user {uid} to {user}.")
        return '', 204

    @error_handler
    @jwt_required
    def delete(self, uid):
        log.info(f"DELETE to {request.url}")
        client = service.find_client()
        # authorization:
        if not policies.check_sysadmin(client):
            raise ForbiddenError(client.username)

        user = service.find_user(uid)
        service.delete_entity(user)
        return '', 204


class UserPassword(Resource):
    @error_handler
    @jwt_required
    def put(self, uid):
        log.info(f"PUT to {request.url}")
        client = service.find_client()
        user = service.find_user(uid)
        # authorization
        if not (policies.check_component(component=client, permissions=[])
                or policies.check_sysadmin(client)
                or client == user):
            raise ForbiddenError(uid=client.id)

        json_data = request.get_json()
        if not json_data:
            raise PortalAPIError(400, message='No data provided for user update.')
        data = password_change_schema.load(json_data)[0]

        if not data.get('new_password'):
            raise PortalAPIError(400, 'missing new password at password change')

        if client == user and not user.is_admin:
            if not data.get('old_password'):
                raise PortalAPIError(400, 'missing old password for password update')
            old_password = data['old_password']
            if not user.verify_password(old_password):
                raise IncorrectPasswordError()

        user.set_password(data['new_password'])
        service.write_entity(user)


class UserList(Resource):
    @error_handler
    @jwt_required
    def get(self):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        # authorization
        if not policies.check_sysadmin(client):
            raise ForbiddenError(uid=client.id)

        course_id = request.args.get('course')
        group_id = request.args.get('group')
        if group_id and not course_id:
            raise PortalAPIError(400, "Invalid filter combination at /get/users: missing course id.")

        users = service.find_users_filtered(course_id=course_id, group_id=group_id)
        return users_schema.dump(users)[0], 200

    @error_handler
    @jwt_required
    def post(self):
        log.info(f"POST to {request.url}")
        client = service.find_client()
        # authorization
        if not policies.check_sysadmin(client):
            raise ForbiddenError(uid=client.id)

        json_data = request.get_json()
        if not json_data:
            raise PortalAPIError(400, message='No data provided for user creation.')
        data = user_schema.load(json_data)[0]
        admin = data.get('is_admin') or False
        new_user = User(uco=data['uco'], email=data['email'], username=data['username'], is_admin=admin)
        service.write_entity(new_user)
        log.debug(f"Created user={new_user}")
        return user_schema.dump(new_user)[0], 201


class UserSubmissionList(Resource):
    # needs more tests
    @error_handler
    @jwt_required
    def get(self, uid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        user = service.find_user(uid)
        # authorization
        if not (policies.check_component(component=client, permissions=[])
                or policies.check_sysadmin(client)
                or client == user):
            raise ForbiddenError(uid=client.id)

        course_id = request.args.get('course')
        project_id = request.args.get('project')

        if project_id and not course_id:
            raise PortalAPIError(400, f"Invalid filter combination at /get/users/{uid}/submissions: missing course id.")

        submissions = service.find_submissions_filtered(user=user, course_id=course_id, project_id=project_id)
        return submissions_schema.dump(submissions)[0], 200


class UserRoleList(Resource):
    @error_handler
    @jwt_required
    def get(self, uid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        user = service.find_user(uid)
        # authorization TODO: insufficient?
        if not (policies.check_component(component=client, permissions=[])
                or policies.check_sysadmin(client)
                or client == user):
            raise ForbiddenError(uid=client.id)

        course_id = request.args.get('course')

        roles = service.find_roles_filtered(user=user, course_id=course_id)
        return roles_schema.dump(roles)[0], 200


class UserCourseList(Resource):
    @error_handler
    @jwt_required
    def get(self, uid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        user = service.find_user(uid)
        # authorization
        if not (policies.check_component(component=client, permissions=[])
                or policies.check_sysadmin(client)
                or client == user):
            raise ForbiddenError(uid=client.id)

        return courses_schema.dump(user.courses)[0], 200


class UserGroupList(Resource):
    @error_handler
    @jwt_required
    def get(self, uid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        user = service.find_user(uid)
        # authorization
        if not (policies.check_sysadmin(client) or client == user):
            raise ForbiddenError(uid=client.id)
        course_id = request.args.get('course')

        groups = service.find_groups_filtered(user=user, course_id=course_id)
        return groups_schema.dump(groups)[0], 200


class UserReviewList(Resource):
    @error_handler
    @jwt_required
    def get(self, uid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        user = service.find_user(uid)
        # authorization
        if not (policies.check_component(component=client, permissions=[])
                or policies.check_sysadmin(client)
                or client == user):
            raise ForbiddenError(uid=client.id)

        reviews = service.find_reviews(user)
        return reviews_schema.dump(reviews)[0], 200


class UserEffectivePermissions(Resource):
    @error_handler
    @jwt_required
    def get(self, uid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        user = service.find_user(uid)
        # authorization TODO - check
        if not (policies.check_component(component=client, permissions=[])
                or policies.check_sysadmin(user)
                or client == user):
            raise ForbiddenError(uid=client.id)

        course_id = request.args.get('course')

        permissions = policies.get_effective_permissions(user, course_id)
        return json.dumps(permissions, cls=DateTimeEncoder), 200


users_api.add_resource(UserResource, '/<string:uid>')
users_api.add_resource(UserPassword, '/<string:uid>/password')
users_api.add_resource(UserList, '')
users_api.add_resource(UserSubmissionList, '/<string:uid>/submissions')
users_api.add_resource(UserRoleList, '/<string:uid>/roles')
users_api.add_resource(UserCourseList, '/<string:uid>/courses')
users_api.add_resource(UserGroupList, '/<string:uid>/groups')
users_api.add_resource(UserReviewList, '/<string:uid>/reviews')
users_api.add_resource(UserEffectivePermissions, '/<string:uid>/permissions')
