from portal.database.models import User, Course, Component, Role
from portal.service import service
from portal.tools.logging import log


def check_component(component: Component, permissions: list, course: Course=None) -> bool:
    """Checks component permissions for a course

    Args:
        component(Component): Component instance
        course(Course): Course instance
        permissions(list): List of permissions required

    Returns(bool): True if checks has passed

    """
    # TODO
    if not isinstance(component, Component):
        return False
    return True


def check_user(user: User, course: Course, permissions: list) -> bool:
    """Checks a user's permissions based on his roles in the course

    Args:
        user(User): User instance
        course(Course): Course instance
        permissions(list): Permissions list

    Returns(bool): True if checks pass

    """
    if not isinstance(user, User):
        return False
    if check_sysadmin(user):
        return True
    roles = user.get_roles_in_course(course=course)

    return all(__evaluate_permission_for_roles(roles, perm) for perm in permissions)


def check_sysadmin(user: User) -> bool:
    """Checks whether the user is sysadmin

    Args:
        user(User): User instance

    Returns(bool): True if checks pass
    """
    if not isinstance(user, User):
        return False
    return user.is_admin


def __evaluate_permission_for_role(role: Role, permission: str, default=False) -> bool:
    """Evaluates permission for a role

    If the permission doesn't exist, False is returned.
    Args:
        role(Role): the role to check
        permission(str): Permission name
        default(bool): Default behaviour when the permission does not exist

    Returns(bool): True if permission is True in the role

    """
    return getattr(role, permission, default=default)


def __evaluate_permission_for_roles(roles: list, permission: str) -> bool:
    """Evaluates permissions for all a list of roles
    Args:
        roles(list): Role list
        permission(str): Permission name

    Returns(bool): True if any of the roles allows the permission

    """
    return any(__evaluate_permission_for_role(role, permission) for role in roles)


def get_effective_permissions(user: User, course_id: str) -> dict:

    course = service.find_course(course_id, throws=False)
    result = {}
    if course is None:
        for c in user.courses:
            result[c.id] = __get_effective_permissions_for_course(user, c)
    else:
        result[course.id] = __get_effective_permissions_for_course(user, course)

    log.debug(f"Found effective permissions for user {user}: {result}")
    return result


def __get_effective_permissions_for_course(user: User, course: Course):
    log.debug(f'getting effective permissions for course {course}')
    result = {}
    course_roles = user.get_roles_in_course(course)
    for role in course_roles:
        log.debug(f'getting effective permissions for role {role}')
        permissions = role.permissions
        for key, value in vars(permissions).items():
            if not key.startswith("_") and key not in ['id', 'role_id', 'role', 'description', 'created_at', 'updated_at']:
                result[key] = result.get(key) or value
    return result
